////////////////////////////////////////////////////////////
// MainDialog.h - 
//
// Copyright (C) Alax.Info, 2006
// http://alax.info
//
// A permission to use the source code is granted as long as reference to 
// source website http://alax.info is retained.

#pragma once

#include <atlstr.h>
#include <atlcrack.h>
#include <atlmisc.h>
#include <atlcoll.h>
#include <atldlgs.h>
#include "resource.h"
#include "SampleObject.h"

////////////////////////////////////////////////////////////
// COlePropertyFrameDialogT, COlePropertyFrameDialog

template <typename T, typename _Base = CWindow, typename _WinTraits = CFrameWinTraits>
class COlePropertyFrameDialogT :
	public CWindowImpl<T, _Base, _WinTraits>
{
	typedef COlePropertyFrameDialogT<T, _Base> COlePropertyFrameDialog;

public:

BEGIN_MSG_MAP_EX(COlePropertyFrameDialogT<T>)
	//ATLTRACE(atlTraceUI, 4, _T("uMsg 0x%x, wParam 0x%08x, lParam 0x%08x\n"), uMsg, wParam, lParam);
END_MSG_MAP()

public:

	////////////////////////////////////////////////////////
	// CPageClassIdentifierArray

	class CPageClassIdentifierArray :
		public CAtlArray<CLSID>
	{
	public:
	// CPageClassIdentifierArray
		SIZE_T Append(const CAtlArray<CLSID>& PageClassIdentifierArray)
		{
			return __super::Append(PageClassIdentifierArray);
		}
		VOID Append(const CLSID* pPageClassIdentifiers, SIZE_T nPageClassIdentifierCount)
		{
			for(SIZE_T nIndex = 0; nIndex < nPageClassIdentifierCount; nIndex++)
				ATLVERIFY(Add(pPageClassIdentifiers[nIndex]) >= 0);
		}
		VOID AppendUnique(const CLSID* pPageClassIdentifiers, SIZE_T nPageClassIdentifierCount)
		{
			for(SIZE_T nIndex = 0; nIndex < nPageClassIdentifierCount; nIndex++)
			{
				const CLSID& PageClassIdentifier = pPageClassIdentifiers[nIndex];
				if(!Find(PageClassIdentifier))
					ATLVERIFY(Add(PageClassIdentifier) >= 0);
			}
		}
		VOID RemoveAbsent(const CLSID* pPageClassIdentifiers, SIZE_T nPageClassIdentifierCount)
		{
			for(SIZE_T nArrayIndex = GetCount() - 1; nArrayIndex + 1 > 0; nArrayIndex--)
			{
				const CLSID& ArrayPageClassIdentifier = GetAt(nArrayIndex);
				BOOL bFound = FALSE;
				for(SIZE_T nIndex = 0; nIndex < nPageClassIdentifierCount; nIndex++)
				{
					const CLSID& PageClassIdentifier = pPageClassIdentifiers[nIndex];
					if(IsEqualCLSID(ArrayPageClassIdentifier, PageClassIdentifier))
					{
						bFound = TRUE;
						break;
					}
				}
				if(!bFound)
					RemoveAt(nArrayIndex);
			}
		}
		BOOL Find(const CLSID& PageClassIdentifier, SIZE_T* pnIndex = NULL, SIZE_T nStartIndex = 0)
		{
			for(SIZE_T nIndex = nStartIndex; nIndex < GetCount(); nIndex++)
				if(IsEqualCLSID(GetAt(nIndex), PageClassIdentifier))
				{
					if(pnIndex)
						*pnIndex = nIndex;
					return TRUE;
				}
			return FALSE;
		}
		SIZE_T Find(const CLSID& PageClassIdentifier, SIZE_T nStartIndex)
		{
			SIZE_T nIndex;
			if(!Find(PageClassIdentifier, &nIndex, nStartIndex))
				return (SIZE_T) -1;
			return nIndex;
		}
	};

private:
	OCPFIPARAMS m_OleCreatePropertyFrameIndirectParameters;
	CStringW m_sCaption;
	CAtlArray<CComPtr<IUnknown> > m_ObjectArray;
	CPageClassIdentifierArray m_PageClassIdentifierArray;
	static HHOOK g_Hook;

	VOID InitializeOleCreatePropertyFrameIndirectParameters() throw()
	{
		ZeroMemory(&m_OleCreatePropertyFrameIndirectParameters, sizeof m_OleCreatePropertyFrameIndirectParameters);
		m_OleCreatePropertyFrameIndirectParameters.cbStructSize = sizeof m_OleCreatePropertyFrameIndirectParameters;
		INT nOffset = GetSystemMetrics(SM_CYFIXEDFRAME) + GetSystemMetrics(SM_CYCAPTION);
		m_OleCreatePropertyFrameIndirectParameters.x = nOffset;
		m_OleCreatePropertyFrameIndirectParameters.y = nOffset;
		m_OleCreatePropertyFrameIndirectParameters.lcid = LOCALE_USER_DEFAULT;
	}
	static LRESULT CALLBACK CallWndProcRetHook(INT nCode, WPARAM wParam, LPARAM lParam) throw()
	{
		CWPRETSTRUCT* pStruct = (CWPRETSTRUCT*) lParam;
		//ATLTRACE(atlTraceUI, 4, _T("nCode %d, wParam %d, lParam 0x%08x { hwnd 0x%08x, message 0x%x, ... }\n"), nCode, wParam, lParam, pStruct ? pStruct->hwnd : (HWND) -1, pStruct ? pStruct->message : -1);
		// NOTE: It appears to be too early to subclass at WM_NCCREATE/WM_CREATE as the frame would abort creation
		if(nCode == HC_ACTION && pStruct)
		{
			if(pStruct->message == WM_INITDIALOG && !(CWindow(pStruct->hwnd).GetStyle() & WS_CHILD))
			{
				T* pT = (T*) ModuleHelper::ExtractCreateWndData();
				ATLASSERT(pT);
				ATLASSERT(!pT->m_hWnd);
				ATLVERIFY(UnhookWindowsHookEx(g_Hook));
				g_Hook = NULL;
				ATLTRACE(atlTraceUI, 3, _T("pStruct->hwnd 0x%08x\n"), pStruct->hwnd);
				ATLASSERT(CWindow(pStruct->hwnd).IsWindow());
				if(pT->SubclassWindow(pStruct->hwnd))
				{
					// NOTE: Re-call WM_INITDIALOG handler for subclasses window
					LRESULT lResult = 0;
					pT->ProcessWindowMessage(pT->m_hWnd, pStruct->message, pStruct->wParam, pStruct->lParam, lResult, 0);
				} else
					ATLTRACE2(atlTraceUI, 1, _T("Failed to subclass OLE property frame\n"));
			}
		}
		return CallNextHookEx(g_Hook, nCode, wParam, lParam);
	}

protected:
	OCPFIPARAMS& GetParameters() throw()
	{
		return m_OleCreatePropertyFrameIndirectParameters;
	}

public:
// COlePropertyFrameDialogT
	COlePropertyFrameDialogT() throw()
	{
		InitializeOleCreatePropertyFrameIndirectParameters();
	}
	COlePropertyFrameDialogT(LPCTSTR pszCaption)
	{
		InitializeOleCreatePropertyFrameIndirectParameters();
		SetCaption(pszCaption);
	}
	const OCPFIPARAMS& GetParameters() const throw()
	{
		return m_OleCreatePropertyFrameIndirectParameters;
	}
	VOID SetCaption(LPCTSTR pszCaption)
	{
		m_sCaption = CT2CW(pszCaption);
		m_OleCreatePropertyFrameIndirectParameters.lpszCaption = m_sCaption;
	}
	VOID SetObjects(IUnknown** ppUnknowns, SIZE_T nUnknownCount)
	{
		m_ObjectArray.RemoveAll();
		for(SIZE_T nIndex = 0; nIndex < nUnknownCount; nIndex++)
			ATLVERIFY(m_ObjectArray.Add(ppUnknowns[nIndex]) >= 0);
		m_OleCreatePropertyFrameIndirectParameters.cObjects = (ULONG) m_ObjectArray.GetCount();
		m_OleCreatePropertyFrameIndirectParameters.lplpUnk = &m_ObjectArray.GetData()->p;
	}
	VOID SetObject(IUnknown* pUnknown)
	{
		ATLASSERT(pUnknown);
		SetObjects(&pUnknown, 1);
	}
	VOID SetPages(const CLSID* pPageClassIdentifiers, SIZE_T nPageClassIdentifierCount)
	{
		m_PageClassIdentifierArray.RemoveAll();
		m_PageClassIdentifierArray.Append(pPageClassIdentifiers, nPageClassIdentifierCount);
		m_OleCreatePropertyFrameIndirectParameters.cPages = (ULONG) m_PageClassIdentifierArray.GetCount();
		m_OleCreatePropertyFrameIndirectParameters.lpPages = m_PageClassIdentifierArray.GetData();
	}
	VOID SetPages(const CAtlArray<CLSID>& PageClassIdentifierArray)
	{
		m_PageClassIdentifierArray.RemoveAll();
		m_PageClassIdentifierArray.Append(PageClassIdentifierArray);
		m_OleCreatePropertyFrameIndirectParameters.cPages = (ULONG) m_PageClassIdentifierArray.GetCount();
		m_OleCreatePropertyFrameIndirectParameters.lpPages = m_PageClassIdentifierArray.GetData();
	}
	VOID SetObjectPages(IUnknown* pUnknown)
	{
		CComHeapPtr<CLSID> pPageClassIdentifiers;
		SIZE_T nPageClassIdentifierCount = 0;
		CComQIPtr<ISpecifyPropertyPages> pSpecifyPropertyPages = pUnknown;
		if(pSpecifyPropertyPages)
		{
			CAUUID Pages = { 0, NULL };
			HRESULT nGetPagesResult = pSpecifyPropertyPages->GetPages(&Pages);
			if(FAILED(nGetPagesResult))
				AtlThrow(nGetPagesResult);
			pPageClassIdentifiers = Pages.pElems;
			nPageClassIdentifierCount = Pages.cElems;
		}
		SetPages(pPageClassIdentifiers, nPageClassIdentifierCount);
	}
	SIZE_T SetObjectPages(IUnknown** ppUnknowns, SIZE_T nUnknownCount, BOOL bAllPages = FALSE)
	{
		CPageClassIdentifierArray PageClassIdentifierArray;
		for(SIZE_T nUnknownIndex = 0; nUnknownIndex < nUnknownCount; nUnknownIndex++)
		{
			CComHeapPtr<CLSID> pPageClassIdentifiers;
			SIZE_T nPageClassIdentifierCount = 0;
			CComQIPtr<ISpecifyPropertyPages> pSpecifyPropertyPages = ppUnknowns[nUnknownIndex];
			if(pSpecifyPropertyPages)
			{
				CAUUID Pages = { 0, NULL };
				HRESULT nGetPagesResult = pSpecifyPropertyPages->GetPages(&Pages);
				if(FAILED(nGetPagesResult))
					AtlThrow(nGetPagesResult);
				pPageClassIdentifiers.Attach(Pages.pElems);
				nPageClassIdentifierCount = Pages.cElems;
			}
			// NOTE: bAllPages instructs to make union/intersection of objects' pages
			if(bAllPages)
				PageClassIdentifierArray.AppendUnique(pPageClassIdentifiers, nPageClassIdentifierCount);
			else
				if(nUnknownIndex == 0)
					PageClassIdentifierArray.Append(pPageClassIdentifiers, nPageClassIdentifierCount);
				else
					PageClassIdentifierArray.RemoveAbsent(pPageClassIdentifiers, nPageClassIdentifierCount);
		}
		SetPages(PageClassIdentifierArray);
		return m_PageClassIdentifierArray.GetCount();
	}
	SIZE_T SetObjectPages()
	{
		return SetObjectPages(&m_ObjectArray.GetData()->p, m_ObjectArray.GetCount());
	}
	INT_PTR DoModal(HWND hParentWindow = GetActiveWindow())
	{
		T* pT = static_cast<T*>(this);
		ATLASSERT(m_OleCreatePropertyFrameIndirectParameters.cObjects);
		ATLASSERT(m_OleCreatePropertyFrameIndirectParameters.cPages);
		m_OleCreatePropertyFrameIndirectParameters.hWndOwner = hParentWindow;
		ATLASSERT(!g_Hook);
		g_Hook = SetWindowsHookEx(WH_CALLWNDPROCRET, CallWndProcRetHook, ModuleHelper::GetModuleInstance(), GetCurrentThreadId());
		if(!g_Hook)
			AtlThrowLastWin32();
		ModuleHelper::AddCreateWndData(&pT->m_thunk.cd, pT);
		HRESULT nOleCreatePropertyFrame = OleCreatePropertyFrameIndirect(&m_OleCreatePropertyFrameIndirectParameters);
		if(g_Hook)
		{
			ATLVERIFY(UnhookWindowsHookEx(g_Hook));
			g_Hook = NULL;
		}
		if(FAILED(nOleCreatePropertyFrame))
			AtlThrow(nOleCreatePropertyFrame);
		return IDOK;
	}

// CWindowImpl
	LRESULT DefWindowProc()
	{
		return __super::DefWindowProc();
	}
	LRESULT DefWindowProc(UINT nMessage, WPARAM wParam, LPARAM lParam)
	{
		// NOTE: Since we hook the frame right after WM_INITDIALOG and re-calling T's
		//       subclasses map once again after the hook, we have to disacrd default 
		//       processing of WM_INITDIALOG message
		if(nMessage == WM_INITDIALOG)
			return 0;
		return __super::DefWindowProc(nMessage, wParam, lParam);
	}
};

// ASSU: Only one UI thread per application (not thread safe)
template <typename T, typename _Base, typename _WinTraits>
__declspec(selectany) HHOOK COlePropertyFrameDialogT<T, _Base, _WinTraits>::g_Hook = NULL;

class COlePropertyFrameDialog :
	public COlePropertyFrameDialogT<COlePropertyFrameDialog>
{
public:
// COlePropertyFrameDialog
};

////////////////////////////////////////////////////////////
// CMyOlePropertyFrameDialog

class CMyOlePropertyFrameDialog :
	public COlePropertyFrameDialogT<CMyOlePropertyFrameDialog>
{
public:

BEGIN_MSG_MAP_EX(CMyOlePropertyFrameDialog)
	CHAIN_MSG_MAP(COlePropertyFrameDialogT<CMyOlePropertyFrameDialog>)
	MSG_WM_INITDIALOG(OnInitDialog)
END_MSG_MAP()

public:
// CMyOlePropertyFrameDialog

// Window message handlers
	LRESULT OnInitDialog(HWND, LPARAM)
	{
		ATLVERIFY(CenterWindow());
		return 0;
	}
};

////////////////////////////////////////////////////////////
// CMainDialog

class CMainDialog : 
	public CDialogImpl<CMainDialog>
{
public:
	enum { IDD = IDD_MAINDLG };

BEGIN_MSG_MAP_EX(CMainDialog)
	MSG_WM_INITDIALOG(OnInitDialog)
	COMMAND_ID_HANDLER_EX(IDOK, OnOk)
	COMMAND_ID_HANDLER_EX(IDCANCEL, OnCancel)
END_MSG_MAP()

public:
// CMainDialog

// Window message handlers
	LRESULT OnInitDialog(HWND, LPARAM)
	{
		ATLVERIFY(CenterWindow());
		SetIcon(AtlLoadIconImage(IDR_MAINFRAME, GetSystemMetrics(SM_CXICON), ::GetSystemMetrics(SM_CYICON)), TRUE);
		SetIcon(AtlLoadIconImage(IDR_MAINFRAME, GetSystemMetrics(SM_CXSMICON), ::GetSystemMetrics(SM_CYSMICON)), FALSE);
		return TRUE;
	}
	LRESULT OnOk(UINT, INT nIdentifier, HWND)
	{
		// NOTE: CSampleObject is a sample object with aattached property page 
		//       (standard ATL/MFC's color property page)
		CComPtr<IUnknown> pSampleObjectUnknown;
		ATLVERIFY(SUCCEEDED(CSampleObject::CreateInstance(&pSampleObjectUnknown)));
		CMyOlePropertyFrameDialog Dialog;
		Dialog.SetObject(pSampleObjectUnknown);
		// NOTE: This instructs dialog to query for proeprty pages from assigned object(s)
		Dialog.SetObjectPages();
		INT_PTR nDialogResult = Dialog.DoModal(m_hWnd);
		nDialogResult;
		//ATLVERIFY(EndDialog(nIdentifier));
		return 0;
	}
	LRESULT OnCancel(UINT, INT nIdentifier, HWND)
	{
		ATLVERIFY(EndDialog(nIdentifier));
		return 0;
	}
};
